##############################################################################
#
# Copyright (c) 2002 Ingeniweb SARL
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""
$$
"""
__author__  = ''
__docformat__ = 'restructuredtext'

# Python imports
import os
import Globals
from Acquisition import aq_base
from StringIO import StringIO
from types import StringType, UnicodeType

# Zope imports
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from OFS.Image import File, Image
from OFS.SimpleItem import SimpleItem
from ZPublisher.Iterators import filestream_iterator
from ComputedAttribute import ComputedAttribute
#from Shared.DC.ZRDB.TM import TM

# CMF imports
from Products.CMFCore.utils import getToolByName

# Archetypes imports
from Products.Archetypes.Storage import StorageLayer
from Products.Archetypes.interfaces.field import IObjectField
from Products.Archetypes.interfaces.base import IBaseUnit

# Products imports
from FileUtils import copy_file, move_file


class VirtualData:

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return self.getData()
    
    def __len__(self):
        return len(self.getData())
    
    __call__ = __str__
    
    def getData(self):
        """Get data from filesystem"""
        
        # Get data from filesystem
        virtual_file = StringIO()
        value = ''
        
        try:
            copy_file(self.path, virtual_file)
            virtual_file.seek(0)
            value = virtual_file.getvalue()
        finally:
            virtual_file.close()
        
        return value
    
    data = ComputedAttribute(getData, 1)

InitializeClass(VirtualData)

class VirtualBinary(VirtualData):

    def __init__(self, path, filename, mimetype, size):
        VirtualData.__init__(self, path)
        self.filename = filename
        self.mimetype = mimetype
        self.size = size
        
    def __len__(self):
        return self.size
    
    def download(self, request):
        
        response = request.RESPONSE
        response.setHeader('Content-Disposition',
                           'attachment; filename="%s"' % self.filename)
        response.setHeader('Content-Type', self.mimetype)
        response.setHeader('Content-Length', self.size)
        return filestream_iterator(self.path, mode='rb')
    
    def get_size(self):
        return len(self)
        
    def getSize(self):
        return len(self)

    def getContentType(self):
        return self.mimetype
    
    content_type = ComputedAttribute(getContentType, 1)
    
    def getFilename(self):
        return self.filename

InitializeClass(VirtualBinary)

class VirtualFile(VirtualBinary, File):
    
    __allow_access_to_unprotected_subobjects__ = 1
    
    def __init__(self, path, filename, mimetype, size):
        VirtualBinary.__init__(self, path, filename, mimetype, size)
    
InitializeClass(VirtualFile)

class VirtualImage(VirtualBinary, Image):
    
    __allow_access_to_unprotected_subobjects__ = 1
    
    def __init__(self, path, filename, mimetype, size, width, height):
        VirtualBinary.__init__(self, path, filename, mimetype, size)
        self.width = width
        self.height = height
    
InitializeClass(VirtualImage)


# ####
# Storage info
# ####

class FSSInfo(SimpleItem):
    """FileSystemStorageInfo Base class. Used for string data"""
    
    security = ClassSecurityInfo()
    
    def __init__(self, uid):
        self.update(uid)
    
    security.declarePrivate('update')
    def update(self, uid):
        self.uid = uid
    
    security.declarePrivate('getUID')
    def getUID(self):
        return self.uid
    
    security.declarePrivate('setUID')
    def setUID(self, uid):
        self.uid = uid
    
    security.declarePrivate('getValue')
    def getValue(self, path):
        return str(VirtualData(path))

InitializeClass(FSSInfo)

class FSSFileInfo(FSSInfo):
    """FileSystemStorageInfo File class. Used for file data"""
    
    security = ClassSecurityInfo()
    
    def __init__(self, uid, title, size, mimetype):
        self.update(uid, title, size, mimetype)
    
    security.declarePrivate('update')
    def update(self, uid, title, size, mimetype):
        FSSInfo.update(self, uid)
        self.title = title
        self.size = size
        self.mimetype = mimetype

    security.declarePrivate('getTitle')
    def getTitle(self):
        return self.title
    
    security.declarePrivate('setTitle')
    def setTitle(self, title):
        self.title = title
    
    security.declarePrivate('getSize')
    def getSize(self):
        return self.size
    
    security.declarePrivate('setSize')
    def setSize(self, size):
        self.size = size
    
    security.declarePrivate('getMimetype')
    def getMimetype(self):
        return self.mimetype
    
    security.declarePrivate('setMimetype')
    def setMimetype(self, mimetype):
        self.mimetype = mimetype
        
    security.declarePrivate('getValue')
    def getValue(self, path):
        return VirtualFile(path, self.title, self.mimetype, self.size)

InitializeClass(FSSFileInfo)

class FSSImageInfo(FSSFileInfo):
    """FileSystemStorageInfo Image class. Used for image data"""
    
    security = ClassSecurityInfo()
    
    def __init__(self, uid, title, size, mimetype, width, height):
        self.update(uid, title, size, mimetype, width, height)
    
    security.declarePrivate('update')
    def update(self, uid, title, size, mimetype, width, height):
        FSSFileInfo.update(self, uid, title, size, mimetype)
        self.width = width
        self.height = height

    security.declarePrivate('getWidth')
    def getWidth(self):
        return self.width
    
    security.declarePrivate('setWidth')
    def setWidth(self, width):
        self.width = width
        
    security.declarePrivate('getHeight')
    def getHeight(self):
        return self.height
    
    security.declarePrivate('setHeight')
    def setHeight(self, height):
        self.height = height
        
    security.declarePrivate('getValue')
    def getValue(self, path):
        return VirtualImage(path, self.title, self.mimetype, self.size, self.width, self.height)

InitializeClass(FSSImageInfo)

# Keep it for compatibility
class FileSystemStorageInfo(FSSFileInfo):
      pass
      
InitializeClass(FileSystemStorageInfo)


# ####
# Storage
# ####

class FileSystemStorage(StorageLayer):
    
    __implements__ = StorageLayer.__implements__
    
    security = ClassSecurityInfo()
    
    security.declarePrivate('getFSSInfoVarname')
    def getFSSInfoVarname(self, name):
        """ """
        
        return '%s_filesystemstorage_info' % name
    
    security.declarePrivate('getFSSInfo')
    def getFSSInfo(self, name, instance, **kwargs):
        """ """
        
        info_varname = self.getFSSInfoVarname(name)
        return getattr(aq_base(instance), info_varname, None)
        
    security.declarePrivate('setFSSInfo')
    def setFSSInfo(self, name, instance, value, **kwargs):
        """ """
        
        info_varname = self.getFSSInfoVarname(name)
        field = self.getField(name, instance, **kwargs)
        
        # Check types
        uid = instance.UID()
        mimetype = kwargs.get('mimetype', getattr(value, 'content_type', None))
        
        if mimetype is not None and hasattr(value, 'get_size'):
           
            size = value.get_size()
            title = kwargs.get('filename', getattr(value, 'filename', getattr(value, 'title', name)))
            width = getattr(value, 'width', None)
            height = getattr(value, 'height', None)
            
            if width is not None and height is not None:
                if type(title) not in (StringType, UnicodeType):
                    # Make sure title is a string (Fix for image thumbs)
                    title = name
                    
                # Like Image field
                info = FSSImageInfo(uid, title, size, mimetype, width, height)
            else:
                # Like File field
                info = FSSFileInfo(uid, title, size, mimetype)
        else:
            # Other 
            info = FSSInfo(uid)
        
        # Make sure, we have deleted old info
        if hasattr(aq_base(instance), info_varname):
            delattr(instance, info_varname)
      
        # Store in attributes
        setattr(instance, info_varname, info)
    
    security.declarePrivate('getField')
    def getField(self, name, instance, **kwargs):
        """Get field"""
        
        return kwargs.get('field', instance.getField(name))
    
    security.declarePrivate('getBackupPath')
    def getBackupPath(self, name, instance, **kwargs):
        """
        """
        
        path = default_path = os.path.join(Globals.INSTANCE_HOME, 'var')
        
        try:
            fss_tool = getToolByName(instance, 'portal_fss')
            path = fss_tool.getBackupPath()
        except:
            # No tool installed
            path = default_path
        
        return path
    
    security.declarePrivate('getBackupFilename')
    def getBackupFilename(self, name, instance, **kwargs):
        """
        This is the same as filename but it uses ".bak" extension.
        """
        
        return '%s.bak' % self.getFilename(name, instance, **kwargs)
    
    security.declarePrivate('getBackupFilePath')
    def getBackupFilePath(self, name, instance, **kwargs):
        """Get backup path of file
        """
        
        path = self.getBackupPath(name, instance, **kwargs)
        filename = self.getBackupFilename(name, instance, **kwargs)
        
        return os.path.join(path, filename)

    security.declarePrivate('getPath')
    def getPath(self, name, instance, **kwargs):
        """
        """
        
        path = default_path = os.path.join(Globals.INSTANCE_HOME, 'var')
        
        try:
            fss_tool = getToolByName(instance, 'portal_fss')
            path = fss_tool.getStoragePath()
        except:
            # No tool installed
            path = default_path
        
        return path
    
    security.declarePrivate('getFilename')
    def getFilename(self, name, instance, **kwargs):
        """
        """
        
        uid = kwargs.get('uid', instance.UID())
        return '%s_%s' % (uid, name)

    security.declarePrivate('getFilePath')
    def getFilePath(self, name, instance, **kwargs):
        """Get path of file
        """
        
        path = self.getPath(name, instance, **kwargs)
        filename = self.getFilename(name, instance, **kwargs)
        
        return os.path.join(path, filename)
    
    security.declarePrivate('get')
    def get(self, name, instance, **kwargs):
        """ """
        
        path = self.getFilePath(name, instance, **kwargs)
        backup_path = self.getBackupFilePath(name, instance, **kwargs)
        info = self.getFSSInfo(name, instance)
        
        if not os.path.exists(path) and os.path.exists(backup_path):
            # Maybe it's an undo
            # Move backup file
            move_file(backup_path, path)
        
        if info is None:
            return ''
        
        value = info.getValue(path)
        
        return value
        
    security.declarePrivate('set')
    def set(self, name, instance, value, **kwargs):
        """ """
        
        field = self.getField(name, instance, **kwargs)
        path = self.getFilePath(name, instance)
        
        # Remove acquisition wrappers
        value = aq_base(value)
        
        # Wrap value
        if IObjectField.isImplementedBy(value):
            new_value = value.getRaw(self.instance)
        if IBaseUnit.isImplementedBy(value):
            new_value = value.getRaw()
        elif isinstance(value, File):
            new_value = value.data
        else:
            new_value = str(value)
        
        # copy data to filesystem
        virtual_file = StringIO(new_value)
        
        try:
            copy_file(virtual_file, path)
        finally:
            virtual_file.close()
            
        # Create File System Storage Info
        self.setFSSInfo(name, instance, value, **kwargs)


    security.declarePrivate('unset')
    def unset(self, name, instance, **kwargs):
        """ """

        path = self.getFilePath(name, instance, **kwargs)

        if os.path.exists(path):
            # Delete file so move to backup
            backup_path = self.getBackupFilePath(name, instance, **kwargs)
            move_file(path, backup_path)

    security.declarePrivate('initializeInstance')
    def initializeInstance(self, instance, item=None, container=None):
        pass
    
    security.declarePrivate('cleanupInstance')
    def cleanupInstance(self, instance, item=None, container=None):
        pass
        
    security.declarePrivate('initializeField')
    def initializeField(self, instance, field):
        """ """

        name = field.getName()
        path = self.getFilePath(name, instance)
        info = self.getFSSInfo(name, instance)
        uid = instance.UID()
        old_uid = uid
        
        if info is not None:
            old_uid = info.getUID()
        
        if uid != old_uid:
            # Copy/Paste
            old_path = self.getFilePath(name, instance, uid=old_uid)
            
            if os.path.exists(old_path):
                copy_file(old_path, path)
      
            # Change UID in info
            info.setUID(uid)
            
    security.declarePrivate('cleanupField')
    def cleanupField(self, instance, field, **kwargs):
        """ """
        is_deleted = not getattr(instance, '_v_cp_refs', False)
        
        if is_deleted:
            name = field.getName()
            self.unset(name, instance, **kwargs)
